/*
 * Copyright (c) 2010 CNRS/LAAS
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/**
 ** niutNiTaskCodels.cc
 **
 ** Codels called by execution task niutNiTask
 **
 ** Author: Matthieu Herrb
 ** Date: December 2010
 **
 **/

#include <portLib.h>

#include <XnOpenNI.h>
#include <XnCodecIDs.h>
#include <XnCppWrapper.h>

#include "server/niutHeader.h"

#ifndef SAMPLE_XML_PATH
#define SAMPLE_XML_PATH "/usr/share/openni/lib/SamplesConfig.xml"
#endif

xn::Context niutContext;
xn::DepthGenerator niutDepthGenerator;
xn::UserGenerator niutUserGenerator;

XnBool needPose = FALSE;
XnChar pose[20] = "";

int initDone = 0;

/**
 * Callback: new user detected
 */
void XN_CALLBACK_TYPE
User_NewUser(xn::UserGenerator& generator, XnUserID nId, void *pCookie)
{
	if (SDI_F->verbose) 
		printf("new user %d\n", nId);
	if (needPose)
		niutUserGenerator.GetPoseDetectionCap().
			StartPoseDetection(pose, nId);
	else
		niutUserGenerator.GetSkeletonCap().
			RequestCalibration(nId, TRUE);
}

/**
 * Callback: lost user
 */
void XN_CALLBACK_TYPE
User_LostUser(xn::UserGenerator& generator, XnUserID nId, void *pCookie)
{
	NIUT_HUMAN_LIST *list = &(SDI_F)->list;
	int i;

	if (SDI_F->verbose) 
		printf("lost user %d\n", nId);
	for (i = 0; i < list->num; i++)
		if (list->users[i].id == nId)
		     list->users[i].state = NIUT_NO_TRACKING;

}

/**
 * Callback: detected a pose
 */
void XN_CALLBACK_TYPE
UserPose_PoseDetected(xn::PoseDetectionCapability& capability,
		      const XnChar *strPos, XnUserID nId, void *pCookie)
{
	if (SDI_F->verbose)
		printf("pose %s detected for user %d\n", pose, nId);
	niutUserGenerator.GetPoseDetectionCap().StopPoseDetection(nId);
	niutUserGenerator.GetSkeletonCap().RequestCalibration(nId, TRUE);
}

/**
 * Callback: started calibration
 */
void XN_CALLBACK_TYPE
UserCalibration_CalibrationStart(xn::SkeletonCapability& capability,
				 XnUserID nId, void*pCookie)
{
	if (SDI_F->verbose) 
		printf("Calibration started for user %d\n", nId);
}

/**
 * Callback: finished calibration
 */
void XN_CALLBACK_TYPE
UserCalibration_CalibrationEnd(xn::SkeletonCapability& capability,
				 XnUserID nId, XnBool success, void *pCookie)
{
	if (success) {
		if (SDI_F->verbose)
			printf("calibration complete, "
			       "start tracking user %d\n", nId);
		niutUserGenerator.GetSkeletonCap().StartTracking(nId);
	} else {
		if (SDI_F->verbose)
			printf("calibration failed for user %d\n", nId);
		User_NewUser(niutUserGenerator, nId, pCookie);
	}
}

/**
 * Update one joint position
 */
void
updateJoint(int i, const char *name, XnSkeletonJoint joint)
{
	NIUT_HUMAN_LIST *list = &(SDI_F->list);
	XnUserID id = list->users[i].id;
	XnSkeletonJointPosition p;

	niutUserGenerator.GetSkeletonCap().GetSkeletonJointPosition(id,
								    joint, p);
	list->users[i].skeleton.joint[joint].confidence = p.fConfidence;
	list->users[i].skeleton.joint[joint].position.x = p.position.X/1000.0;
	list->users[i].skeleton.joint[joint].position.y = p.position.Y/1000.0;
	list->users[i].skeleton.joint[joint].position.z = p.position.Z/1000.0;
	
#if 0	
	if (p.fConfidence < 0.5)
		return;
	printf("-- %s %.2lf %.2lf %.2lf\n", name, p.position.X/1000.0,
	       p.position.Y/1000.0,
	       p.position.Z/1000.0);
#endif
}


/**
 * Update skeleton position
 */
void
updateSkeleton(int i)
{
	updateJoint(i, "HEAD", XN_SKEL_HEAD);
	updateJoint(i, "NECK", XN_SKEL_NECK);

	updateJoint(i, "LEFT_SHOULDER", XN_SKEL_LEFT_SHOULDER);
	updateJoint(i, "LEFT_ELBOW", XN_SKEL_LEFT_ELBOW);
	updateJoint(i, "LEFT_HAND", XN_SKEL_LEFT_HAND);

	updateJoint(i, "RIGHT_SHOULDER", XN_SKEL_RIGHT_SHOULDER);
	updateJoint(i, "RIGHT_ELBOW", XN_SKEL_RIGHT_ELBOW);
	updateJoint(i, "RIGHT_HAND", XN_SKEL_RIGHT_HAND);

	updateJoint(i, "TORSO", XN_SKEL_TORSO);

	updateJoint(i, "LEFT_HIP", XN_SKEL_LEFT_HIP);
	updateJoint(i, "LEFT_KNEE", XN_SKEL_LEFT_KNEE);
	updateJoint(i, "LEFT_FOOT", XN_SKEL_LEFT_FOOT);

	updateJoint(i, "RIGHT_HIP", XN_SKEL_RIGHT_HIP);
	updateJoint(i, "RIGHT_KNEE", XN_SKEL_RIGHT_KNEE);
	updateJoint(i, "RIGHT_FOOT", XN_SKEL_RIGHT_FOOT);
}

/**
 * Handle data
 */
void
handleData(void)
{
	NIUT_HUMAN_LIST *list = &(SDI_F->list);
	int wd = list->watch_dog;
	xn::SceneMetaData sceneMD;
	xn::DepthMetaData depthMD;
	XnUInt16 nUsers = 15;
	XnUserID allUsers[15];
	XnPoint3D com;
	int i;

	niutContext.WaitAndUpdateAll();

	niutDepthGenerator.GetMetaData(depthMD);
	niutUserGenerator.GetUserPixels(0, sceneMD);

	niutUserGenerator.GetUsers(allUsers, nUsers);
	
	memset(list,0,sizeof(NIUT_HUMAN_LIST));

	list->watch_dog = wd;
	list->num = nUsers;

	for (i = 0; i < nUsers; i++) {
		//printf("-> user %d\n", allUsers[i]);
		niutUserGenerator.GetCoM(allUsers[i], com);
		list->users[i].id = allUsers[i];
		if (niutUserGenerator.GetSkeletonCap().
		    IsTracking(allUsers[i])) {
			list->users[i].state = NIUT_TRACKING;
			//printf("tracking %d\n", allUsers[i]);
			updateSkeleton(i);
		} else if (niutUserGenerator.GetSkeletonCap().
			   IsCalibrating(allUsers[i])) {
			list->users[i].state = NIUT_CALIBRATE;
			//printf("calibrating %d\n", allUsers[i]);
		} else {
			list->users[i].state = NIUT_POSE_SEARCH;
			//printf("looking for pose %d\n", allUsers[i]);
		}
	}
}


/*------------------------------------------------------------------------
 *
 * niut_update  --  1st codel of permanent activity (called BEFORE other activities)
 *
 * Description: 
 * 
 * Reports:      OK
 * 
 * Returns:    OK or ERROR
 */

STATUS
niut_update(int *report)
{
	if (initDone) 
	{
	  handleData();

	  // The counter is incremented to
	  // keep the reader informed of a potential death
	  (SDI_F->list.watch_dog)++;
	  if(SDI_F->list.watch_dog > 100000)
	    SDI_F->list.watch_dog = 0;
	}

	return OK;
}
/*------------------------------------------------------------------------
 * Init
 *
 * Description: 
 *
 * Reports:      OK
 *              S_niut_NI_ERROR
 */

/* niut_init  -  codel EXEC of Init
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
niut_init(int *report)
{
	XnStatus retval = XN_STATUS_OK;
	XnCallbackHandle userCallbacks, calibrationCallbacks, poseCallbacks;

	printf("\nUsing %s\n", SAMPLE_XML_PATH);

	/* Initializations */
	retval = niutContext.InitFromXmlFile(SAMPLE_XML_PATH);
	if (retval != XN_STATUS_OK) {
		printf("InitFromXml failed: %s\n", xnGetStatusString(retval));
		*report = S_niut_NI_ERROR;
		return ETHER;
	}
	retval = niutContext.FindExistingNode(XN_NODE_TYPE_DEPTH,
					      niutDepthGenerator);
	if (retval != XN_STATUS_OK) {
		printf("find depth generator failed: %s\n",
		       xnGetStatusString(retval));
		*report = S_niut_NI_ERROR;
		return ETHER;
	}
	retval = niutContext.FindExistingNode(XN_NODE_TYPE_USER,
					      niutUserGenerator);
	if (retval != XN_STATUS_OK) {
		printf("user generator not found, creating it\n");
		retval = niutUserGenerator.Create(niutContext);
		if (retval != XN_STATUS_OK) {
			printf("create user generator failed: %s\n",
			       xnGetStatusString(retval));
			*report = S_niut_NI_ERROR;
			return ETHER;
		}
	}

	if (!niutUserGenerator.IsCapabilitySupported(XN_CAPABILITY_SKELETON)) {
		printf("Supplied user generator doesnt support skeleton\n");
		*report = S_niut_NI_ERROR;
		return ETHER;
	}
	/* Register callbacks */
	niutUserGenerator.RegisterUserCallbacks(User_NewUser, User_LostUser,
						NULL, userCallbacks);
	niutUserGenerator.GetSkeletonCap().
		RegisterCalibrationCallbacks(UserCalibration_CalibrationStart,
					     UserCalibration_CalibrationEnd,
					     NULL, calibrationCallbacks);

	if (niutUserGenerator.GetSkeletonCap().NeedPoseForCalibration()) {
		printf("need pose for calibration\n");
		needPose = TRUE;
		if (!niutUserGenerator.
		    IsCapabilitySupported(XN_CAPABILITY_POSE_DETECTION)) {
			printf("pose required, but not supported\n");
			*report = S_niut_NI_ERROR;
			return ETHER;
		}
		niutUserGenerator.GetPoseDetectionCap().
			RegisterToPoseCallbacks(UserPose_PoseDetected,
						NULL, NULL, poseCallbacks);
		niutUserGenerator.GetSkeletonCap().GetCalibrationPose(pose);
	}
	/* Track full body */
	niutUserGenerator.GetSkeletonCap().
		SetSkeletonProfile(XN_SKEL_PROFILE_ALL);

	/* Go */
	retval = niutContext.StartGeneratingAll();
	if (retval != XN_STATUS_OK) {
		printf("start generators failed: %s\n",
		       xnGetStatusString(retval));
		*report = S_niut_NI_ERROR;
		return ETHER;
	}
	SDI_F->list.watch_dog = -1;
	SDI_F->verbose = 0;
	initDone = 1;
	return ETHER;
}


/*------------------------------------------------------------------------
 * Stop
 *
 * Description: 
 *
 * Reports:      OK
 */

/* niut_stop  -  codel EXEC of Stop
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
niut_stop(int *report)
{
	*report = OK;
	if (initDone)
		niutContext.Shutdown();

	memset(&niutDataStrId->list,0,sizeof(NIUT_HUMAN_LIST));
	initDone = 0;
	return ETHER;
}
